/* ez-node.h: General purpose node.

   Copyright (C) 2010,2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef _EZ_NODE_H
#define _EZ_NODE_H

#include <stdint.h>

typedef struct ez_node ez_node;

/**
 * A list/tree/set node.
 *
 * All objects stored in the list/tree/set must begin with this
 * structure.  It does not need initialisation.
 *
 * Use the appropriate node accessors to navigate.  They are only
 * valid iff the node is a member of a container.
 *
 * The order of each is important.
 */
struct ez_node {
	union {
		struct {
			struct ez_node * volatile succ;
			struct ez_node * volatile pred;
		} list;
		struct {
			intptr_t link[2];
		} tree;
		struct {
			struct ez_node *next;
			unsigned int hash;
		} set;
	} u;
};

/**
 * Direction of a tree node link.
 */
enum ez_node_link_t {
	EZ_LINK_LEFT = 0,
	EZ_LINK_RIGHT = 1
};

typedef unsigned int (*ez_hash_fn)(const void *entry);
typedef int (*ez_equal_fn)(const void *a, const void *b);
typedef int (*ez_cmp_fn)(const void *a, const void *b);
typedef void (*ez_free_fn)(void *);

/**
 * Retrieve the list node successor.
 *
 * The end of the list is indicated by node.next == NULL, not by node
 * == NULL.
 */
static __inline__ void *ez_node_succ(void *n) {
	return (void *)((struct ez_node *)n)->u.list.succ;
}

/**
 * Retrieve the list node predecessorr.
 *
 * The start of the list is indicated by node.prev == NULL, not by node
 * == NULL.
 */
static __inline__ void *ez_node_pred(void *n) {
	return (void *)((struct ez_node *)n)->u.list.pred;
}

/**
 * Retrieve the left child tree node.
 */
static __inline__ void *ez_node_left(void *node) {
	return (void *)(((struct ez_node *)node)->u.tree.link[0] & ~3);
}

/**
 * Retrieve the right child tree node.
 */
static __inline__ void *ez_node_right(void *node) {
	return (void *)(((struct ez_node *)node)->u.tree.link[1] & ~3);
}

/**
 * Retrieve either the left or right child node.
 */
static __inline__ void *ez_node_link(void *node, enum ez_node_link_t i) {
	return (void *)(((struct ez_node *)node)->u.tree.link[i] & ~3);
}

/**
 * Retrieve the next set hash chain bucket.
 */
static __inline__ void *ez_node_next(void *node) {
	return ((struct ez_node *)node)->u.set.next;
}

/**
 * Retrieve the set hash bucket hash.
 */
static __inline__ unsigned int ez_node_hash(void *node) {
	return ((struct ez_node *)node)->u.set.hash;
}

static void *ez_node_succ(void *n) __attribute__ ((always_inline));
static void *ez_node_pred(void *n) __attribute__ ((always_inline));
static void *ez_node_left(void *node) __attribute__((always_inline));
static void *ez_node_right(void *node) __attribute__((always_inline));
static void *ez_node_link(void *node, enum ez_node_link_t i) __attribute__((always_inline));
static void *ez_node_next(void *node) __attribute__((always_inline));
static unsigned int ez_node_hash(void *node) __attribute__((always_inline));

#endif
